-- number of revealed tiles required for 1 exp point
local iTiles1ExpPoint = 8;

local iTiles1ExpPointCur = iTiles1ExpPoint;
local tScouts = {};
local tScoutTypes = {};
local tNumTilesRevealed = {};
local tHumans = {};
local tAIexplorerTypes = {};
local tAIscoutsMoved = {};
local iAIExplorersMoved = 0;
local iSurveyPolicy = GameInfo.Policies["POLICY_SURVEY"].Index;


local function TableLength(T)
  local count = 0
  for _ in pairs(T) do count = count + 1 end
  return count
end


local function CacheScoutTypes()
	for row in GameInfo.Units() do
		if row.PromotionClass == 'PROMOTION_CLASS_RECON' then
			tScoutTypes[row.Index] = true;
		end
	end
end


local function CacheAIexplorerTypes()
	for row in GameInfo.UnitAiInfos() do
		if row.AiType == 'UNITAI_EXPLORE' then
			local iType = GameInfo.Units[row.UnitType].Index;
			tAIexplorerTypes[iType] = true;
		end
	end
end


local function SaveRevealedTiles(iPlayer)
	local pVisibility = PlayersVisibility[iPlayer];
	local iExploredTiles = pVisibility:GetNumRevealedHexes();
	tNumTilesRevealed[iPlayer] = iExploredTiles;
	-- print('SaveRevealedTiles ' , iPlayer, iExploredTiles); 
end


local function CheckSurveyPolicy(iPlayerID)
	local pCulture = Players[iPlayerID]:GetCulture();

	if pCulture:IsPolicyActive(iSurveyPolicy) then
		iTiles1ExpPointCur = math.floor(iTiles1ExpPoint/2);	
	else
		iTiles1ExpPointCur = iTiles1ExpPoint;	
	end
end


local function GiveExpToScout (iPlayerID, iUnitID, iX, iY)
	if not tHumans[iPlayerID] then
		return -- AI reveals tiles before this fires
	end
	
	local iTilesRevealed = PlayersVisibility[iPlayerID]:GetNumRevealedHexes();
	local iSavedTilesRevealed = tNumTilesRevealed[iPlayerID];
	local iTilesRevealedThisMove = iTilesRevealed - iSavedTilesRevealed;
	-- print('OnUnitMoved ' , iPlayerID, pUnit:GetName()); 
	-- print('OnUnitMoved iTilesRevealedThisMove', iTilesRevealedThisMove, 'iTilesRevealed', iTilesRevealed, 'iSavedTilesRevealed', iSavedTilesRevealed); 
	if iTilesRevealedThisMove > 0 then
		local iUnitTilesRevealed = tScouts[iPlayerID][iUnitID];
		if iUnitTilesRevealed then
			local pUnit = UnitManager.GetUnit(iPlayerID, iUnitID);
			iUnitTilesRevealed = iUnitTilesRevealed + iTilesRevealedThisMove;
			local iExp = math.floor(iUnitTilesRevealed/iTiles1ExpPointCur);
			local iRemainder = iUnitTilesRevealed - iExp*iTiles1ExpPointCur;
			-- print('scout Moved Tiles Revealed', iTilesRevealedThisMove);
			if iExp > 0 then
				-- print('give',iExp, 'exp, Remainder', iRemainder);
				pUnit:GetExperience():ChangeExperience(iExp);
				pUnit:SetProperty('SGEE_TilesRevealed', iRemainder);
				tScouts[iPlayerID][iUnitID] = iRemainder;
			else
				-- print('save revealed scout tiles', iUnitTilesRevealed);
				pUnit:SetProperty('SGEE_TilesRevealed', iUnitTilesRevealed);
				tScouts[iPlayerID][iUnitID] = iUnitTilesRevealed;
			end
		end
		tNumTilesRevealed[iPlayerID] = iTilesRevealed;
	end
end


-- fires when settler founds city
local function OnUnitOperationAdded(iPlayerID, iUnitID, _, iType)
	-- print('OnUnitOperationAdded', iPlayerID, GameInfo.UnitOperations[iType].OperationType); 
	if tHumans[iPlayerID] then
		local pUnit = UnitManager.GetUnit(iPlayerID, iUnitID);
		-- local iUnitType = pUnit:GetType();
		-- local pVisibility = PlayersVisibility[iPlayerID];
		-- will already be updated
		-- local iExploredTiles = pVisibility:GetNumRevealedHexes();
		if pUnit and tScouts[iPlayerID][iUnitID] then
		else
			-- print('OnUnitOperationAdded', pUnit:GetName()); 
			SaveRevealedTiles(iPlayerID);
		end
	end
end


--  does not fire when game loads
local function OnPlayerTurnStarted(iPlayerID)
	local pVisibility = PlayersVisibility[iPlayerID];
	local iExplored = pVisibility:GetNumRevealedHexes();
	CheckSurveyPolicy(iPlayerID);
	SaveRevealedTiles(iPlayerID);
	-- print('OnPlayerTurnStarted ', iPlayerID, iExplored); 
end


local function OnPolicyChanged(iPlayerID, iPolicyID, bEnacted)
	-- print('OnPolicyChanged', GameInfo.Policies[iPolicyID].PolicyType, bEnacted)
	CheckSurveyPolicy(iPlayerID);
end


local function OnPlayerTurnDeactivated(iPlayerID)
	local iScoutsMoved = TableLength(tAIscoutsMoved);

	if not tHumans[iPlayerID] and iScoutsMoved > 0 then
		local pVisibility = PlayersVisibility[iPlayerID];
		local iRevealed = pVisibility:GetNumRevealedHexes();
		local iRevealedLastTurn = tNumTilesRevealed[iPlayerID];
		local iRevealedThisTurn = iRevealed - iRevealedLastTurn;
		local iRevealedPerUnit = math.floor(iRevealedThisTurn/iAIExplorersMoved);
		-- print('OnPlayerTurnDeactivated RevealedLastTurn', iPlayerID, iRevealedLastTurn); 
		-- print('OnPlayerTurnDeactivated RevealedThisTurn', iRevealedThisTurn); 
		-- print('OnPlayerTurnDeactivated RevealedPerUnit', iRevealedPerUnit); 
		
		if iRevealedPerUnit > 0 then
			for iUnitID, pUnit in pairs(tAIscoutsMoved) do
				local iUnitTilesRevealed = tScouts[iPlayerID] and tScouts[iPlayerID][iUnitID];
				if not iUnitTilesRevealed then
					iUnitTilesRevealed = 0;
				end
				iUnitTilesRevealed = iUnitTilesRevealed + iRevealedPerUnit;
				local iExp = math.floor(iUnitTilesRevealed/iTiles1ExpPointCur);
				local iRemainder = iUnitTilesRevealed - iExp*iTiles1ExpPointCur;
				-- print('AI scout Tiles Revealed', iUnitTilesRevealed);
				if iExp > 0 then
					-- print('give',iExp, 'exp, Remainder', iRemainder);
					-- local iRandomChance = Game.GetRandNum(101, 'SGEE');
					-- if iChanceAIscoutGetsExp >= iRandomChance then
						pUnit:GetExperience():ChangeExperience(iExp);
						pUnit:SetProperty('SGEE_TilesRevealed', iRemainder);
					-- end
					tScouts[iPlayerID][iUnitID] = iRemainder;
				else
					-- print('save revealed tiles', iUnitTilesRevealed);
					pUnit:SetProperty('SGEE_TilesRevealed', iUnitTilesRevealed);
					tScouts[iPlayerID][iUnitID] = iUnitTilesRevealed;
				end
			end
		end
	end
	tAIscoutsMoved = {};
	iAIExplorersMoved = 0;
end


local function OnUnitMovementPointsCleared(iPlayerID, iUnitID)
	if not tHumans[iPlayerID] then
		local pUnit = UnitManager.GetUnit(iPlayerID, iUnitID);
		if tAIexplorerTypes[pUnit:GetType()] then
			iAIExplorersMoved = iAIExplorersMoved + 1;
		end
		if tScouts[iPlayerID][iUnitID] then
			tAIscoutsMoved[iUnitID] = pUnit;
			-- print('Scout Movement Points Cleared', pUnit:GetName(), pUnit:GetOperationType());
		end
	end
end


local function OnUnitAddedToMap(iPlayerID, iUnitID, iX, iY)

	local pUnit = UnitManager.GetUnit(iPlayerID, iUnitID);
	if tScoutTypes[pUnit:GetType()] then
		-- print('Scout Added To Map ', iPlayerID, iUnitID, iX, iY); 
		local iUnitTilesRevealed = pUnit:GetProperty('SGEE_TilesRevealed') or 0;
		tScouts[iPlayerID][pUnit:GetID()] = iUnitTilesRevealed;
	end
end


local function Initialize()
	CacheScoutTypes();
	CacheAIexplorerTypes();

	for iPlayerID, pPlayer in ipairs(Players) do
		if pPlayer:WasEverAlive() then
			CheckSurveyPolicy(iPlayerID);

			if not tScouts[iPlayerID] then
				tScouts[iPlayerID] = {};
			end 
			if not tNumTilesRevealed[iPlayerID] then
				tNumTilesRevealed[iPlayerID] = {};
			end 
			for i, pUnit in pPlayer:GetUnits():Members() do
				if tScoutTypes[pUnit:GetType()] then
					local iUnitTilesRevealed = pUnit:GetProperty('SGEE_TilesRevealed') or 0;
					tScouts[iPlayerID][pUnit:GetID()] = iUnitTilesRevealed;
				end
			end
			if pPlayer:IsHuman() then
				tHumans[iPlayerID] = true;
			end
			SaveRevealedTiles(iPlayerID);
		end
	end
	Events.UnitMoved.Add(GiveExpToScout);
	Events.UnitPromoted.Add(GiveExpToScout); -- Promotion may give +1 sight
	GameEvents.PlayerTurnStarted.Add(OnPlayerTurnStarted);
	Events.UnitAddedToMap.Add(OnUnitAddedToMap);
	Events.PlayerTurnDeactivated.Add(OnPlayerTurnDeactivated);
	Events.UnitMovementPointsCleared.Add(OnUnitMovementPointsCleared);
	Events.UnitOperationAdded.Add(OnUnitOperationAdded);
	GameEvents.PolicyChanged.Add(OnPolicyChanged);
	print('SGEE LUA LOADED'); 
end
Events.LoadGameViewStateDone.Add(Initialize)

--[[
local function OnUnitSelectionChanged(iPlayerID, unitID, hexI, hexJ, hexK, bSelected, bEditable)
	local pVisibility = PlayersVisibility[iPlayerID];
	local iExplored = pVisibility:GetNumRevealedHexes();
	SaveRevealedTiles(iPlayerID);
	print('OnUnitSelectionChanged ', iPlayerID, iExplored); 
end
-- Events.UnitSelectionChanged.Add( OnUnitSelectionChanged );

local function OnUnitActivityChanged(iOwner, iUnitID, x, y, eReason, bVisibleToLocalPlayer)
	local pPlayer = Players[iOwner];
	local pUnits = pPlayer:GetUnits();
	local pUnit = pUnits:FindID(iUnitID);
	local pVisibility = PlayersVisibility[iOwner];
	local iExploredCount = pVisibility:GetNumRevealedHexes();
	if pUnit then
		print('OnUnitActivityChanged', pUnit:GetName(), iExploredCount); 
	end
end

local function OnUnitOperationDeactivated(iPlayer, iUnitID, iOperationHash)
	local pPlayer = Players[iPlayer];
	local pUnits = pPlayer:GetUnits();
	local pUnit = pUnits:FindID(iUnitID);
	local pVisibility = PlayersVisibility[iPlayer];
	local iExploredCount = pVisibility:GetNumRevealedHexes();
	if pUnit then
		print('OnUnitOperationDeactivated', pUnit:GetName(), iExploredCount); 
	end
end

local function OnUnitOperationSegmentComplete(iPlayer, iUnitID, iOperationHash)
	local pPlayer = Players[iPlayer];
	local pUnits = pPlayer:GetUnits();
	local pUnit = pUnits:FindID(iUnitID);
	local pVisibility = PlayersVisibility[iPlayer];
	local iExploredCount = pVisibility:GetNumRevealedHexes();
	if pUnit then
		print('OnUnitOperationSegmentComplete', pUnit:GetName(), iExploredCount); 
	end
end

--]]